import waba.ui.*;
import waba.fx.*;
import waba.sys.*;
import waba.util.*;

public class Bench extends MainWindow
{
   class TestClass
   {
      public int field1 = 70;
      public int field2 = 80;

      public int method1(int value)
      {
         return (field1 % value) + field2 / 2;
      }
      
      public int method2(int value)
      {
         return (field2 % value) + field1 / 2;
      }
   }

   Graphics g;
   Font boldFont = new Font("Helvetica", Font.BOLD, 12);
   Font plainFont = new Font("Helvetica", Font.PLAIN, 12);

   public int field1 = 70;
   public int field2 = 80;

   public Bench()
   {
   }
   public void onStart()
   {
      runTests();
   }
   public void runTests()
   {
      Sound.beep();      
      g = new Graphics(this);
      FontMetrics fm = new FontMetrics(plainFont, this);
      int lineHeight = fm.getHeight() + fm.getLeading();

      // clear
      // clear area
      g.setColor(255, 255, 255);
      g.fillRect(0, 0, 160, 160);
      g.setColor(0,0,0);
      int y = 40;
      int start = Vm.getTimeStamp();
      for (int i = 0; i < 8000; i++)
      {
         switch (Vm.random(0) % 2)
         {
            case 0: g.drawRect(Vm.random(0) % 150,Vm.random(0) % 150,10,10); break; 
            case 1: g.drawLine(Vm.random(0) % 150,Vm.random(0) % 150,Vm.random(0) % 150,Vm.random(0) % 150); break; 
         }
      }
      int end = Vm.getTimeStamp();
      g.setColor(255, 255, 255);
      g.fillRect(0, 0, 160, 160);
      g.setColor(0, 0, 0);
      g.drawText("native meth./graph: " + (end - start) + " millis", 10, y);
      y += lineHeight;

      // loop test
      start = Vm.getTimeStamp();
      int value = 100;
      for (int i = 0; i < 100000; i++)
      {
         value = value / 7;
         if (value == 0)
            value = i / 3;
      }
      end = Vm.getTimeStamp();
      g.drawText("Loop: " + (end - start) + " millis", 10, y);

      // field test
      TestClass test = new TestClass();
      start = Vm.getTimeStamp();
      value = 100;
      for (int i = 0; i < 10000; i++)
      {
         value += test.field1 + test.field2;
         test.field1 = i;
         test.field2 = i;
         value = value % 100;
      }
      end = Vm.getTimeStamp();
      y += lineHeight;
      g.drawText("Field: " + (end - start) + " millis", 10, y);

      // method test
      start = Vm.getTimeStamp();
      value = 100;
      for (int i = 1; i < 10000; i++)
      {
         value = test.method1(i);
         value += test.method2(i);
         value = value % 100;
      }
      end = Vm.getTimeStamp();
      y += lineHeight;
      g.drawText("Method: " + (end - start) + " millis", 10, y);

      // array test
      byte array[] = new byte[30];
      start = Vm.getTimeStamp();
      for (int i = 0; i < 10000; i++)
      for (int j = 0; j < 30; j++)
      array[j] = (byte)(array[j] + (j / 2));
      end = Vm.getTimeStamp();
      y += lineHeight;
      g.drawText("Array: " + (end - start) + " millis", 10, y);

      // string test
      String strings[] = new String[30];
      for (int i = 0; i < 30; i++)
         strings[i] = "Num" + i;
      start = Vm.getTimeStamp();
      for (int i = 0; i < 100; i++)
      {
         String z = "All";
         for (int j = 0; j < 30; j++)
            z += strings[j];
      }
      end = Vm.getTimeStamp();
      y += lineHeight;
      g.drawText("String: " + (end - start) + " millis", 10, y);
      
      y += lineHeight;
      g.drawText("wait 40 secs to close", 20, y);
      Vm.setTrace(true,300,0,4);
      Sound.beep();
      Vm.sleep(40000);
      exit(0);
   }

   public void onPaint(int x, int y, int width, int height)
   {
   }
}
